from copy import deepcopy
from os.path import join


def get_before_slash(string):
    """
    Get the part of a string before the first slash
    :param str string: String to return the first part of
    :return str part_before_first_slash: Part of the string before
    the first slash
    """
    return string.split('/')[0]


def get_after_slash(string):
    """
    Get the part of a string after the first slash
    :param str string: String to return the part after the first
    slash of
    :return str part_after_first_slash: Part of the string after
    the first slash
    """
    return '/'.join(string.split('/')[1:])


def inflate_key_value_pair(k, v):
    """
    From a key and value, return a dict representing that key
    as a nested dictionary where the levels are determined
    by the slashes in the key path.

    Example:
    key = 'some/key', value = 'value' ->
    {'some': {'key': 'value'}}
    :param str k: key of the key value pair. If it contains
    slashes those will determined the nested levels.
    :param str v: value of the key value pair. This will be the
    deepest nested value
    :return dict nested_dict: nested dict representing the key value pair
    """
    return {
        get_before_slash(k): inflate_key_value_pair(
            get_after_slash(k), v
        )
    } if '/' in k else {k: v}


def dict_merge(first_dict, second_dict):
    """
    Merge two dicts recursively. Similar to dict.update but instead
    merge nested items instead of overwriting the entire value as soon
    as the same key is encountered. Does not mutate the dicts.

    https://gist.github.com/angstwad/bf22d1822c38a92ec0a9

    Example:
    dict 1: {'a': {'b': 1}}
    dict 2: {'a': {'c': 2}}

    Output: {'a': {'b': 1, 'c': 2}}

    Using dict.update twice would overwrite the dict 1 value 'b', this
    function builds a dict at deeper levels (Union).
    :param dict first_dict: A dict to combine with another
    :param dict second_dict: A dict to combine with another
    :return dict combined_dict: The combined dict.
    """
    def merge(dict1, dict2):
        for k, v in dict2.items():
            if k in dict1 and isinstance(dict1[k], dict):
                merge(dict1[k], dict2[k])
            else:
                dict1[k] = dict2[k]
    merged_dict = deepcopy(first_dict)
    merge(merged_dict, second_dict)
    return merged_dict


def loop_dictionary(dictionary, path='', callback=lambda path, k, v: None):
    """
    Loop the dictionary and perform the callback for each value
    :param dict dictionary: dictionary to loop for values
    :param str path: the depth in the dict joined by /
    :param func callback: the callback to perform for each value
    :return None:
    """
    for k, v in dictionary.items():
        if isinstance(v, dict):
            loop_dictionary(v, path=join(path, k), callback=callback)
        else:
            callback(path, k, v)



from consul_kv.api import put_kv, get_kv, get_kv_cas, get_kv_meta, delete_kv, put_kv_txn
from consul_kv.serializer import map_dictionary, dictionary_map
from consul_kv.settings import DEFAULT_KV_ENDPOINT, DEFAULT_REQUEST_TIMEOUT


class Connection(object):
    """
    Client for the consul key value store API
    """
    kv_endpoint = DEFAULT_KV_ENDPOINT
    timeout = DEFAULT_REQUEST_TIMEOUT

    @staticmethod
    def _kv_endpoint_to_txn_endpoint(kv_endpoint):
        """
        Rewrite a key/value endpoint to a txn endpoint for atomic updates
        :param str kv_endpoint: endpoint to replace the kv path for with txn
        :return str txn_endpoint: the url with the kv path replaced with txn
        """
        return kv_endpoint.replace('/v1/kv', '/v1/txn')

    def __init__(self, endpoint=None, timeout=None):
        self.kv_endpoint = endpoint or self.kv_endpoint
        self.txn_endpoint = self._kv_endpoint_to_txn_endpoint(
            self.kv_endpoint
        )
        self.timeout = timeout or DEFAULT_REQUEST_TIMEOUT

    def put(self, k, v, cas=None):
        """
        Put a key value pair at the configured endpoint
        :param str k: key to put
        :param int cas: the cas version number. If None, no cas is used.
        :param str v: value to put
        :return None:
        """
        return put_kv(k, v, cas, endpoint=self.kv_endpoint, timeout=self.timeout)

    def put_mapping(self, mapping):
        """
        Atomically (Txn) put a key/value mapping at the configured endpoint
        :param dict mapping: dict of key/values put
        :return None:
        """
        return put_kv_txn(
            mapping, endpoint=self.txn_endpoint, timeout=self.timeout
        )

    def put_dict(self, dictionary):
        """
        Atomically (Txn) put a dict at the configured endpoint
        :param dict dictionary: dict of nested keys and values
        {'some': {'key1': 'value1', 'key2': 'value2'} ->
        [{'some/key1': 'value1, 'some/key2': 'value2}]
        :return None:
        """
        mapping = map_dictionary(dictionary)
        return self.put_mapping(mapping)

    def get_cas(self, k=None, recurse=False):
        """
        Get a value for a key and use CAS to guard updates against updates
        from multiple clients.
        :param str k: key to get
        :param bool recurse: return nested entries
        :return dict mapping: retrieved key/value mapping
        """
        return get_kv_cas(
            k=k, recurse=recurse, endpoint=self.kv_endpoint,
            timeout=self.timeout
        )

    def get_meta(self, k=None, recurse=False):
        """
        Get the raw un-decoded key value data for a key
        :param str k: key to get
        :param bool recurse: return nested entries
        :return dict: raw API response
        """
        return get_kv_meta(
            k=k, recurse=recurse, endpoint=self.kv_endpoint,
            timeout=self.timeout
        )

    def get(self, k=None, recurse=False):
        """
        Get a value for a key or all values under that key if recursive is specified
        :param str k: key to get
        :param bool recurse: return nested entries
        :return dict mapping: retrieved key/value mapping
        """
        return get_kv(
            k=k, recurse=recurse, endpoint=self.kv_endpoint,
            timeout=self.timeout
        )

    def get_mapping(self, k=None):
        """
        Retrieve a key value mapping from a specified key
        Note: contrary to put_mapping, this method is not atomic
        :param k: The key to get recursively and return as a key value mapping
        :return dict mapping: The retrieved key value mapping
        """
        return self.get(k=k, recurse=True)

    def get_dict(self, k=None):
        """
        Retrieve a dict that represents nested key value entries from the specified
        endpoint. [{'some/key1': 'value1, 'some/key2': 'value2}] ->
        {'some': {'key1': 'value1', 'key2': 'value2'}
        Note: contrary to put_dict, this method is not atomic
        :param k: The key to get recursively and return as a dict
        :return dict dictionary: The retrieved dict of nested keys and values
        """
        mapping = self.get_mapping(k=k)
        return dictionary_map(mapping)

    def delete(self, k=None, recurse=False):
        """
        Delete the specified key or all values under that key if recursive is specified
        :param str k: key to delete
        :param bool recurse: delete nested entries
        :return None:
        """
        return delete_kv(
            k=k, recurse=recurse,
            endpoint=self.kv_endpoint, timeout=self.timeout
        )



import socket
from base64 import b64decode, b64encode
from json import loads, dumps
from logging import getLogger
from os.path import join
from urllib import request
from urllib.parse import urlencode

from consul_kv.settings import DEFAULT_KV_ENDPOINT, DEFAULT_TXN_ENDPOINT

log = getLogger(__name__)


def put_kv(
    k, v, cas=None, endpoint=DEFAULT_KV_ENDPOINT,
    timeout=socket._GLOBAL_DEFAULT_TIMEOUT
):
    """
    Put a key and value to the distributed key value store at the location path
    :param str k: the key to put
    :param int cas: the cas version number. If None, no cas is used.
    :param str v: the value to put
    :param str endpoint: API url to PUT to
    :param int timeout: Seconds before timing out
    :return None:
    """
    encoded = str.encode(str(v))
    params = dict()
    if cas:
        params['cas'] = cas

    url = join(endpoint, k) if k else endpoint

    if params:
        url = "{}/?{}".format(url, urlencode(params))

    req = request.Request(
        url=url, data=encoded, method='PUT'
    )
    with request.urlopen(req, timeout=timeout) as f:
        log.debug("PUT k v pair ({}, {}) to {}: {}, {}".format(
            k, v, url, f.status, f.reason
        ))


def _mapping_to_txn_data(mapping, verb='set'):
    """
    Transform a key value mapping to a list of operations to perform
    inside the atomic transaction.
    :param dict mapping: flat dict of key/values, or key/(value, modification-index, verb)
    :param str verb: The type of operation to perform. See the list of possibilities
    here https://www.consul.io/docs/agent/http/kv.html#txn
    :return list[dict, ..] txn_data: List of dicts describing the operations to perform
    """
    return [
        {
            'KV': {
                'Verb': v[2],
                'Key': k,
                'Value': b64encode(str(v[0]).encode('utf-8')).decode('utf-8'),
                'Index': v[1]
            } if isinstance(v, tuple) else
            {
                'Verb': verb,
                'Key': k,
                'Value': b64encode(str(v).encode('utf-8')).decode('utf-8'),
            }
        } for k, v in mapping.items()
    ]


def put_kv_txn(mapping, endpoint=DEFAULT_TXN_ENDPOINT, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
    """
    Update multiple keys inside a single, atomic transaction.
    The body of the request should be a list of operations to
    perform inside the atomic transaction. Up to 64 operations
    may be present in a single transaction. Each Base64-encoded
    blob of data can not be larger than 512kB.
    https://www.consul.io/docs/agent/http/kv.html
    :param dict mapping: flat dict of key/values put
    :param str endpoint: API url to PUT to. Should be a txn endpoint.
    :param int timeout: Seconds before timing out
    :return None:
    """
    txn_data = _mapping_to_txn_data(mapping, verb='set')
    data = dumps(txn_data).encode('utf-8')
    req = request.Request(
        url=endpoint, data=data, method='PUT',
        headers={'Content-Type': 'application/json'}
    )
    with request.urlopen(req, timeout=timeout) as f:
        log.debug("PUT k v mapping {} to {}: {}, {}".format(
            mapping, endpoint, f.status, f.reason
        ))


def get_kv_builder(postprocessor=lambda x: x):
    """
    Function that returns a function that gets the key value mapping
    from the distributed key value store and postprocesses the result
    before returning the value.
    :param func postprocessor: The function to apply to the result
    before returning it as the result of the returned function
    The default postprocessor simply returns the value as is.
    :return func get_kv_raw: The function that gets the key value
    and applies the passed postprocessor before returning the retrieved
    value
    """
    def get_kv_raw(
        k=None, recurse=False, endpoint=DEFAULT_KV_ENDPOINT,
        timeout=socket._GLOBAL_DEFAULT_TIMEOUT
    ):
        """
        Get the key value mapping from the distributed key value store and
        apply the defined postprocessor from the outer context before
        returning the retrieved value.
        :param str k: key to get
        :param bool recurse: whether or not to recurse over the path and
        retrieve all nested values
        :param str endpoint: API url to get the value from
        :param int timeout: Seconds before timing out
        :return dict|mul mapping: key value mapping, or the return value
        as defined in the specified postprocessor of the outer function.
        """
        url = join(endpoint, k) if k else endpoint
        url = "{}/?recurse".format(url) if recurse else url
        req = request.Request(
            url=url,
            method='GET'
        )
        with request.urlopen(req, timeout=timeout) as r:
            result = loads(r.read().decode('utf-8'))
        return postprocessor(result)
    return get_kv_raw

get_kv_meta = get_kv_builder()

get_kv_cas = get_kv_builder(
    lambda x: {
        r['Key']: (b64decode(r['Value']).decode('utf-8'), r['ModifyIndex'])
        if r['Value'] else None for r in x if r['Key']
    }
)

get_kv = get_kv_builder(
    lambda x: {
        # values are stored base64 encoded in consul, they
        # are decoded before returned by this function.
        r['Key']: b64decode(r['Value']).decode('utf-8')
        if r['Value'] else None for r in x if r['Key']
    }
)


def delete_kv(
    k=None, cas=None, recurse=False, endpoint=DEFAULT_KV_ENDPOINT,
    timeout=socket._GLOBAL_DEFAULT_TIMEOUT
):
    """
    Delete a key from the distributed key value store
    :param str k: the key to delete
    :param int cas: the cas version number. If None, no cas is used.
    :param bool recurse: recurse the path and delete all entries
    :param str endpoint: API url to DELETE
    :param int timeout: Seconds before timing out
    :return:
    """
    params = dict()
    if cas:
        params['cas'] = cas
    if recurse:
        params['recurse'] = recurse

    url = join(endpoint, k) if k else endpoint

    if params:
        url = "{}/?{}".format(url, urlencode(params))

    req = request.Request(
        url=url,
        method='DELETE'
    )
    with request.urlopen(req, timeout=timeout) as f:
        log.debug("DELETEd key {}{}: {} {}".format(
            url,
            ' recursively' if recurse else '',
            f.status,
            f.reason
        ))



from distutils.core import setup
setup(
    name ='consul_kv',
    packages=['consul_kv'],
    version='0.6',
    description='Consul KV client for Python 3',
    author='Rick van de Loo',
    author_email='rickvandeloo@gmail.com',
    url='https://github.com/vdloo/consul-kv',
    keywords=['consul', 'distributed', 'key', 'value'],
    classifiers=[],
)



from os.path import join

from consul_kv.utils import loop_dictionary, inflate_key_value_pair, dict_merge


def map_dictionary(dictionary):
    """
    Map a dictionary to key value pairs where each key is an
    accumulation of all levels of nesting for the value
    :param dict dictionary: to flatten into a k/v mapping
    :return dict mapping: key value mapping
    """
    mapping = dict()

    def add_item_to_mapping(path, k, v):
        mapping.update({join(path, k): v})

    loop_dictionary(
        dictionary,
        callback=add_item_to_mapping
    )
    return mapping


def dictionary_map(mapping):
    """
    Create a dictionary from a mapping of key value pairs where
    each key is represented by nested dicts based on slashes in
    the key path
    :param dict mapping: key value mapping
    :return dict dictionary: k/v mapping represented as a nested dict
    """
    dictionary = dict()

    def add_item_to_dictionary(path, k, v):
        dictionary.update(
            dict_merge(
                dictionary,
                inflate_key_value_pair(
                    join(path, k), v
                )
            )
        )

    loop_dictionary(
        mapping,
        callback=add_item_to_dictionary
    )

    return dictionary
